from django.db import models
from mint import mixins
from mint import exceptions
from mint import serializer
from django.conf import settings
from django.http import QueryDict
import simplejson


class EmptyModel(models.Model):
    class Meta:
        app_label = '_mint'


class Controller(mixins.ListMixin, mixins.DetailMixin, mixins.FieldMixin):
    model = EmptyModel
    serializer = serializer.Serializer

    max_limit = None

    def __init__(self, request):
        self.request = request
        self.original_model = self.model
        self.args = {}
        self.limit = None
        self.offset = None
        self.open_args()
        self.open_limit_and_offset()
        self.serializer_instance = self.serializer(self)

    def open_model(self, idx=None):
        if idx:
            try:
                self.model = self.model.objects.get(pk=idx)
            except self.model.DoesNotExist:
                raise exceptions.HttpNotFound("Could not find %s with id %d" % (self.__class__.__name__, int(idx)))
        else:
            self.model = self.model.objects.all()

    def open_args(self):
        if self.request.method == 'GET':
            self.args = self.request.GET.dict()
        elif self.request.method == 'POST':
            if 'application/json' in self.request.META['CONTENT_TYPE']:
                if len(self.request.body) > 0:
                    self.args = simplejson.loads(self.request.body)
            else:
                self.args = self.request.POST.dict()
        elif self.request.method in ('PUT', 'PATCH', 'DELETE'):
            if 'application/json' in self.request.META['CONTENT_TYPE']:
                if len(self.request.body) > 0:
                    self.args = simplejson.loads(self.request.body)
            else:
                self.args = QueryDict(self.request.body).dict()

    def open_limit_and_offset(self):
        if 'limit' in self.args:
            self.limit = int(self.args['limit'])
        if self.limit and self.max_limit and self.limit > self.max_limit:
            self.limit = self.max_limit
        elif self.max_limit and not self.limit:
            self.limit = self.max_limit
        self.offset = int(self.args.get('offset', 0))

    def has_action(self, action):
        return hasattr(self, action)

    def exec_root(self):
        if self.request.method not in self._list_methods():
            raise exceptions.HttpNotAllowed("Invalid method (%s) for this controller (%s)."
                                            % (self.request.method, self.__class__.__name__))
        return self._list_methods()[self.request.method]()

    def exec_id(self, idx):
        if self.request.method not in self._detail_methods():
            raise exceptions.HttpNotAllowed("Invalid method (%s) for this controller (%s)."
                                            % (self.request.method, self.__class__.__name__))
        self.open_model(idx)
        return self._detail_methods()[self.request.method]()

    def exec_action(self, action, idx=None):
        if not hasattr(self, action):
            raise exceptions.HttpNotAllowed("Invalid action (%s) for this controller (%s)."
                                            % (action, self.__class__.__name__))
        return getattr(self, action)()


from django.contrib.auth import get_user_model
from django.db.models.query import QuerySet
from mint import utils
from mint import exceptions
import collections


class BaseMixin(object):
    def _filter(self, queryset):
        return queryset

    def _slice(self, queryset):
        if self.limit and self.offset:
            return queryset[self.offset:self.offset+self.limit]
        elif self.limit:
            return queryset[:self.limit]
        elif self.offset:
            return queryset[self.offset]
        return queryset

    def _pre_run(self):
        pass

    def _post_run(self, model):
        pass

    def _pre_serialize(self):
        pass

    def _post_serialize(self):
        pass

    def _pre_delete(self, model):
        pass

    def _post_delete(self):
        pass

    def _pre_create(self, model):
        pass

    def _post_create(self, model):
        pass

    def _pre_update(self, model):
        pass

    def _post_update(self, model):
        pass

    def _pre_save(self, model):
        pass

    def _post_save(self, model):
        pass

    def _can_get(self):
        return True

    def _can_list(self):
        return True

    def _can_update(self):
        return True

    def _can_delete(self):
        return False

    def _can_create(self):
        return True

    def _iterate_and_serialize(self, model):
        out = []
        for m in model:
            out.append(self._serialize(m))
        return out

    def _get_serializer_fields(self):
        return []

    def _serialize(self, model):
        self._pre_serialize()
        out = self.serializer_instance.pack(model)
        self._post_serialize()
        return out

    def _get_name(self):
        return self.__class__.__name__

    def _format_plural(self, out, name=None, extra={}):
        if not name:
            name = utils.camel_to_underscore(self._get_name())
        if name[-1:] != "s":
            name = "%ss" % name
        extra[name] = out
        ret = dict(extra, **self._get_return_metadata(self.model))
        return ret

    def _format_singular(self, out, name=None, extra={}):
        if not name:
            name = utils.camel_to_underscore(self._get_name())
        if name[-1:] == "s":
            name = name[:-1]
        extra[name] = out
        ret = dict(extra, **self._get_return_metadata(self.model))
        return ret

    def _return(self, model):
        if isinstance(model, collections.Iterable):
            return self._iterate_and_serialize(model)
        return self._serialize(model)

    def _get_return_metadata(self, model):
        if isinstance(model, QuerySet):
            count = model.count()
        else:
            count = 1
        return {
            'count': count,
            'limit': self.limit,
            'offset': self.offset
        }


class ListMixin(BaseMixin):
    def list(self):
        self._pre_run()
        if not self._can_list():
            raise exceptions.HttpUnauthorized("You cannot list this object")
        self.model = self._slice(self._filter(self.model))
        self._post_run(self.model)
        return self._format_plural(self._return(self.model))

    def create(self):
        self._pre_run()
        if not self._can_create():
            raise exceptions.HttpUnauthorized("You cannot delete this object")
        m = self.original_model()
        for field in self.serializer_instance.required_create_fields:
            if field not in self.args.keys():
                raise exceptions.ParameterError("Parameter '%s' is required." % field)
        for key, value in self.args.items():
            if hasattr(m, key) and key in self.serializer_instance.can_create_fields:
                my_type = m._meta.get_field(key).get_internal_type()
                if my_type in ('DateTimeField', 'DateField'):
                    if value:
                        value = utils.datetime_from_string(value)
                elif hasattr(self.model, 'get_%s_display' % key):
                    values = dict((key, value) for (value, key) in self.model._meta.get_field(key).choices)
                    if value not in values:
                        raise exceptions.ParameterError("Invalid value for '%s' parameter.")
                    value = values[value]
                setattr(m, key, value)
        self._pre_create(m)
        self._pre_save(m)
        m.save()
        self._post_save(m)
        self._post_create(m)
        self._post_run(m)
        return self._format_singular(self._return(m))

    def _put_root(self):
        raise exceptions.HttpNotAllowed("Cannot PUT on root context")

    def _delete_root(self):
        raise exceptions.HttpNotAllowed("Cannot DELETE on root context")

    def _list_methods(self):
        return {
            'GET': self.list,
            'POST': self.create,
            'PUT': self._put_root,
            'DELETE': self._delete_root,
        }


class DetailMixin(BaseMixin):
    def get(self):
        self._pre_run()
        if not self._can_get():
            raise exceptions.HttpUnauthorized("You cannot get this object")
        self._post_run(self.model)
        return self._format_singular(self._return(self.model))

    def _post_by_id(self):
        raise exceptions.HttpNotAllowed("Cannot POST by ID")

    def update(self):
        self._pre_run()
        if not self._can_update():
            raise exceptions.HttpUnauthorized("You cannot update this object")
        self._pre_update(self.model)
        for field in self.serializer_instance.required_update_fields:
            if field not in self.args.keys():
                raise exceptions.ParameterError("Parameter '%s' is required." % field)
        for key, value in self.args.items():
            if hasattr(self.model, key) and key in self.serializer_instance.can_update_fields:
                my_type = self.model._meta.get_field(key).get_internal_type()
                if my_type in ('DateTimeField', 'DateField'):
                    value = utils.datetime_from_string(value)
                elif hasattr(self.model, 'get_%s_display' % key):
                    values = dict((key, value) for (value, key) in self.model._meta.get_field(key).choices)
                    if value not in values:
                        raise exceptions.ParameterError("Invalid value for '%s' parameter.")
                    value = values[value]
                setattr(self.model, key, value)
        self._pre_save(self.model)
        self.model.save()
        self._post_save(self.model)
        self._post_update(self.model)
        self._post_run(self.model)
        return self._format_singular(self._return(self.model))

    def delete(self):
        self._pre_run()
        if not self._can_delete():
            raise exceptions.HttpUnauthorized("You cannot delete this object")
        self._pre_delete(self.model)
        self.model.delete()
        self._post_delete()
        self._post_run(self.model)
        return True

    def _detail_methods(self):
        return {
            'GET': self.get,
            'POST': self._post_by_id,
            'PUT': self.update,
            'DELETE': self.delete,
        }


class FieldMixin(object):
    def has_field(self, name):
        if 'fields' not in self.args:
            return False
        fields = self.args['fields'].split(",")
        return name in fields



from django.conf import settings
from django.core import serializers
from mint import utils
import collections


class Serializer(object):
    return_fields = ()

    can_update_fields = ()
    can_create_fields = ()

    required_update_fields = ()
    required_create_fields = ()

    transformers = {}

    default_transformers = {
        'DateTimeField': utils.string_from_datetime,
        'DateField': utils.string_from_datetime,
        'TimeField': utils.string_from_time,
        'DecimalField': float,
        'FileField': lambda value: "%s%s" % (settings.MEDIA_URL, value),
        'ManyToManyField': lambda value: []

    }

    def __init__(self, controller):
        self.controller = controller

    def pack(self, model):
        fields = serializers.serialize('python', (model, ))[0]['fields']
        out = {}
        out_fields = self.return_fields
        for name, value in fields.items():
            if name in out_fields:
                if hasattr(self, name):
                    relationship = getattr(self, name)
                    out[name] = relationship.serialize(model)
                else:
                    my_type = model._meta.get_field(name).get_internal_type()
                    out[name] = self.transform_field(model, my_type, name, value)
                    if my_type in ['ForeignKey']:
                        key = "%s_id" % name
                        out[key] = value
                        del out[name]
        if 'id' in out_fields:
            out['id'] = model.id
        out['model'] = utils.camel_to_underscore(model.__class__.__name__)
        self.pre_return(model, out)
        return out

    def transform_field(self, model, type, name, value):
        transformers = dict(self.default_transformers, **self.transformers)
        if type in transformers:
            value = transformers[type](value)
        if hasattr(model, 'get_%s_display' % name):
            return getattr(model, 'get_%s_display' % name)()
        else:
            return value

    def pre_return(self, model, out):
        pass


class Relationship(object):
    def __init__(self, serializer_class, field_name, required_field=None):
        self.field_name = field_name
        self.serializer_class = serializer_class
        self.required_field = required_field
        self.serializer_instance = None

    def get_serializer(self):
        if not self.serializer_instance:
            self.serializer_instance = self.serializer_class()
        return self.serializer_instance

    def check_for_field(self):
        if self.required_field:
            return self.get_serializer().controller.has_field(self.required_field)
        return True

    def serialize(self, model):
        raise NotImplementedError("Must be implemented in extending class.")


class ForeignKeyRelationship(Relationship):
    def serialize(self, model):
        if not self.check_for_field():
            return None
        field = getattr(model, self.field_name)
        if field is not None:
            return self.get_serializer().pack(field)
        return field


class ManyToManyFieldRelationship(Relationship):
    def serialize(self, model):
        out = []
        if not self.check_for_field():
            return out
        for model in getattr(model, self.field_name).all():
            out.append(self.get_serializer().pack(model))
        return out


def serialize(controller, serializer_class, model):
    if isinstance(model, collections.Iterable):
        out = []
        for item in model:
            out.append(serializer_class(controller).pack(item))
        return out
    return serializer_class(controller).pack(model)



from distutils.core import setup


setup(
    name='django-mint',
    packages=['mint'],
    version='0.33',
    description='RESTful API for Django',
    author='Michael Bates',
    author_email='michael@mbates.co',
    url='https://github.com/pastapreneur/mint',
    download_url='https://github.com/pastapreneur/mint/tarball/0.33',
    keywords=['django', 'api', 'rest', 'restful'],
    classifiers=[],
    install_requires=['simplejson', 'python-dateutil']
)


from django.conf.urls import url
from mint import views


def get_urlconf(coordinator):
    return [
        url(r'^(?P<ctx>[a-z_]+)/(?P<idx>[0-9]+)/(?P<action>[a-z_]+)/?$', views.with_id, {'coordinator': coordinator}),
        url(r'^(?P<ctx>[a-z_]+)/(?P<idx>[0-9]+)/?$', views.with_id, {'coordinator': coordinator}),
        url(r'^(?P<ctx>[a-z_]+)/(?P<action>[a-z_]+)/?$', views.without_id, {'coordinator': coordinator}),
        url(r'^(?P<ctx>[a-z_]+)/?$', views.without_id, {'coordinator': coordinator}),
    ]


def with_id(request, coordinator, ctx, idx, action=None):
    api = coordinator(request, ctx, idx=idx, action=action)
    return api.run()


def without_id(request, coordinator, ctx, action=None):
    api = coordinator(request, ctx, action=action)
    return api.run()